# -*- coding: utf-8 -*-

import torch
import torch.nn.functional as F
from torch_geometric.data import Data

import networkx as nx
import numpy as np

"""# Utils"""

def __parse_nx_edgelist(graph, layer_ids, intralayer_ids, in_channels):
  graph_layers = []

  # Layers
  for l in layer_ids:
    layer_edges = [(u,v) for u,v,e in graph.edges(data=True) if e['layer'] == l]
    layer = nx.Graph()
    layer.add_edges_from(layer_edges)
    nx.set_node_attributes(layer, values=np.random.rand(layer.number_of_nodes(), in_channels), name='features')
    graph_layers.append(layer)

  # Interlayer edges
  for layer1, layer2 in zip(graph_layers[:-1], graph_layers[1:]):
    interlayer_edges = [(u,v) for u,v,e in graph.edges(data=True) if e['layer'] in intralayer_ids]
    layer1.add_edges_from(interlayer_edges, interlayer=True)
    layer2.add_edges_from(interlayer_edges, interlayer=True)
    # Featuring all nodes (also nodes in the interlayer edges)
    nx.set_node_attributes(layer1, values=np.random.rand(layer.number_of_nodes(), in_channels), name='features')
    nx.set_node_attributes(layer2, values=np.random.rand(layer.number_of_nodes(), in_channels), name='features')

  # Convert to Data
  data, data_idx_mapping = __layers_to_data_list(graph_layers)
  return data, data_idx_mapping, graph_layers

def __layers_to_data_list(graph_layers):
  data_list = []
  data_idx_mapping = []
  enumerate_start = 0
  for layer in graph_layers:
    data, idx_mapping = __graph_to_data(layer)
    data_list.append(data)
    data_idx_mapping.append(idx_mapping)
    enumerate_start += len(layer.nodes())
  return data_list, data_idx_mapping

def __graph_to_data(graph):
  idx_mapping = { node: idx for idx, node in enumerate(graph.nodes()) }
  edge_index = []
  edge_index_inter = []
  for src, dst in graph.edges():
      if src in idx_mapping and dst in idx_mapping:
          edge_index.append((idx_mapping[src], idx_mapping[dst]))
          if graph[src][dst].get('interlayer', False):
              edge_index_inter.append((idx_mapping[src], idx_mapping[dst]))

  edge_index = torch.tensor(edge_index, dtype=torch.long).t().contiguous()
  edge_index_inter = torch.tensor(edge_index_inter, dtype=torch.long).t().contiguous()

  features_list = list(nx.get_node_attributes(graph, 'features').values())
  x = torch.tensor(np.array(features_list), requires_grad=True) ## eventually, remove requires_grad=True

  data = Data(x=x, edge_index=edge_index, edge_index_inter=edge_index_inter)
  data.validate()
  # inverse mapping for key-value
  idx_mapping = {v : k for k, v in idx_mapping.items()}
  return data, idx_mapping

"""
# Reading data
"""

def read_data_from_nx(graph1_layers:nx.Graph, graph2_layers:nx.Graph, layer_ids:[], interlayer_ids:[], in_channels:int):
  data_multi_graph1, graph1_node_labels, _ = __parse_nx_edgelist(graph1_layers, layer_ids, interlayer_ids, in_channels)
  data_multi_graph2, graph2_node_labels, _ = __parse_nx_edgelist(graph2_layers, layer_ids, interlayer_ids, in_channels)
  return data_multi_graph1, data_multi_graph2, graph1_node_labels, graph2_node_labels

def read_data_from_file(source_path:str, target_path:str, layer_ids:[], interlayer_ids:[], in_channels:int, delimiter:str=" "):
  source = nx.read_edgelist( source_path, delimiter=delimiter, data=True )
  target = nx.read_edgelist( target_path, delimiter=delimiter, data=True )
  return read_data_from_nx(source, target, layer_ids, interlayer_ids, in_channels)


# -*- coding: utf-8 -*-

import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn import MessagePassing

"""# GIN Model"""

class GINLayer(MessagePassing):
  def __init__(self, in_channels, out_channels):
    super(GINLayer, self).__init__(aggr='add')
    self.mlp = nn.Sequential(
      nn.Linear(in_channels, out_channels),
      nn.ReLU(),
      nn.Linear(out_channels, out_channels)
    )

  def forward(self, x, edge_index, edge_attr):
    x = x.to(self.mlp[0].weight.dtype)
    return self.propagate(edge_index, x=x, edge_attr=edge_attr)

  def message(self, x_j, edge_attr):
    return x_j

class GIN(nn.Module):
  def __init__(self, in_channels, hidden_channels, out_channels, num_layers):
    super(GIN, self).__init__()
    self.conv1 = GINLayer(in_channels, hidden_channels)
    self.convs = nn.ModuleList()
    for _ in range(num_layers - 2):
      self.convs.append(GINLayer(hidden_channels, hidden_channels))
    self.conv2 = GINLayer(hidden_channels, out_channels)

  def forward(self, x, edge_index, edge_attr):
    x = self.conv1(x, edge_index, edge_attr)
    for conv in self.convs:
      x = conv(x, edge_index, edge_attr)
    x = self.conv2(x, edge_index, edge_attr)
    return x

#

def custom_loss(embeddings_graph1, embeddings_graph2):
  # Calculate cosine similarity
  cosine_sim = F.cosine_similarity(embeddings_graph1, embeddings_graph2)
  # A negative cosine_sim maximizes the cosine similarity
  loss = -cosine_sim.mean()
  return loss


from setuptools import setup

setup(
    name='pymulsim',
    version='0.1.1',    
    description='A package for computing pairwise node similarities between multilayer networks, based on embeddings from Graph Neural Networks.',
    url='https://github.com/pietrocinaglia/pymulsim',
    author='Pietro Cinaglia',
    author_email='cinaglia@unicz.it',
    license='MIT',
    packages=['pymulsim'],
    install_requires=['numpy>=1.23.2',
                      'networkx>=2.8.6',
                      'torch>=2.0.1',
                      'torch_geometric>=2.3.1'            
                      ],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',  
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.8',
    ],
)


# -*- coding: utf-8 -*-

import torch
import torch.nn.functional as F
import numpy as np
import networkx as nx
#
from .GIN import GIN
from .Utils import read_data_from_file, read_data_from_nx

class PyMulSim:

  def __init__(self, verbose=False):
    self.verbose = verbose
    self.channels = 64
    self.epochs = 1
    pass

  def __compute_similarities(self, data_multi_graph1:[], data_multi_graph2:[], graph1_node_labels:[]=None, graph2_node_labels:[]=None):

    num_layers = len(data_multi_graph1)

    # GNN model
    model = GIN(in_channels=self.channels, hidden_channels=self.channels, out_channels=self.channels, num_layers=num_layers)

    if self.verbose: print(model)
    
    output = {}
    #output_old = [] 
    for layer_i in range(0, num_layers):
      if self.verbose: print("-- Processing Layer " + str(layer_i))
      layer_i_graph1 = data_multi_graph1[layer_i]
      layer_i_graph2 = data_multi_graph2[layer_i]

      optimizer = torch.optim.Adam(model.parameters(), lr=0.001)
      for epoch in range(self.epochs):
          model.train()
          optimizer.zero_grad()
          embeddings_layer_i_graph1 = model(layer_i_graph1.x, layer_i_graph1.edge_index, layer_i_graph1.edge_index_inter)
          embeddings_layer_i_graph2 = model(layer_i_graph2.x, layer_i_graph2.edge_index, layer_i_graph2.edge_index_inter)
          loss = F.mse_loss(embeddings_layer_i_graph1, embeddings_layer_i_graph2)
          #loss = custom_loss(embeddings_layer_i_graph1, embeddings_layer_i_graph2)
          loss.backward()
          optimizer.step()
          if self.verbose: print("Epoch %d (loss=%d)" % (epoch, float(loss)))

      with torch.no_grad():
        model.eval()

        if self.verbose: print("--- Node embedding...")
        # Using GNNAlignmentModel for processing aligned features of each layer
        embeddings_layer_i_graph1 = model(layer_i_graph1.x, layer_i_graph1.edge_index, layer_i_graph1.edge_index_inter)
        embeddings_layer_i_graph2 = model(layer_i_graph2.x, layer_i_graph2.edge_index, layer_i_graph2.edge_index_inter)

        # Normalization
        if self.verbose: print("--- Normalization...")
        embeddings_layer_i_graph1 = embeddings_layer_i_graph1 / embeddings_layer_i_graph1.norm(dim=1)[:, None]
        embeddings_layer_i_graph2 = embeddings_layer_i_graph2 / embeddings_layer_i_graph2.norm(dim=1)[:, None]

        # Calculate Jaccard similarity between node neighborhoods
        jaccard_similarities = np.zeros((len(embeddings_layer_i_graph1), len(embeddings_layer_i_graph2)))
        for i, embedding_i in enumerate(embeddings_layer_i_graph1):
            for j, embedding_j in enumerate(embeddings_layer_i_graph2):
                neighbors_i = set(np.where(layer_i_graph1.edge_index[0] == i)[0])
                neighbors_j = set(np.where(layer_i_graph2.edge_index[0] == j)[0])
                union_size = len(neighbors_i.union(neighbors_j))
                if union_size == 0:
                    jaccard_sim = 0.0
                else:
                    jaccard_sim = len(neighbors_i.intersection(neighbors_j)) / union_size
                jaccard_similarities[i, j] = jaccard_sim

        # Print all Jaccard similarities
        for i in range(len(embeddings_layer_i_graph1)):
            for j in range(len(embeddings_layer_i_graph2)):
                node_name_graph1 = graph1_node_labels[layer_i][i]
                node_name_graph2 = graph2_node_labels[layer_i][j]
                similarity = jaccard_similarities[i, j]
                if similarity > 0:
                  output[(node_name_graph1, node_name_graph2)] = similarity
                  # output_old.append([node_name_graph1, node_name_graph2, similarity])
                  if self.verbose: print(f"{node_name_graph1} (Graph1) - {node_name_graph2} (Graph2) : {similarity}")

    return output # np.asarray(output_old), 

  ####
        
  def compute_similarities_from_file(self,source_path:str, target_path:str, layer_ids:[], interlayer_ids:[]):
    data_multi_graph1, data_multi_graph2, graph1_node_labels, graph2_node_labels = read_data_from_file(source_path, target_path, layer_ids, interlayer_ids, self.channels)
    return self.__compute_similarities(data_multi_graph1, data_multi_graph2, graph1_node_labels, graph2_node_labels)

  def compute_similarities_from_nx(self,graph1_layers:nx.Graph, graph2_layers:nx.Graph, layer_ids:[], interlayer_ids:[]):
    data_multi_graph1, data_multi_graph2, graph1_node_labels, graph2_node_labels = read_data_from_nx(graph1_layers, graph2_layers, layer_ids, interlayer_ids, self.channels)
    return self.__compute_similarities(data_multi_graph1, data_multi_graph2, graph1_node_labels, graph2_node_labels)
  